#!/usr/bin/perl -w

# 
# i-scream central monitoring system
# http://www.i-scream.org.uk
# Copyright (C) 2000-2002 i-scream
# 
# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License
# as published by the Free Software Foundation; either version 2
# of the License, or (at your option) any later version.
# 
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
# 
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
#

# -----------------------------------------------------------
# i-scream graph generation script
# http://www.i-scream.org.uk
#
# Generates graphs from rrd databases for i-scream data.
#
# $Author: tdb $
# $Id: graph.pl,v 1.12 2003/03/10 11:34:43 tdb Exp $
#------------------------------------------------------------

## TODO
# possibly make more configurable?
#  -- allow configurable periods of graphs
#  -- comments, types, etc

my($version) = '$Id: graph.pl,v 1.12 2003/03/10 11:34:43 tdb Exp $';

$| = 1;

use strict;
use Getopt::Std;
use RRDs;

# define variables that will be read from the config
# nb. keep this insync with the config file!
use vars qw{ 
    $imgdir $rrddir                     
    $maxrrdage $maximgage $deleterrds $deleteimgs
    $hex_slash $hex_underscore  
    $rrdstep $retry_wait
    $verbose $quiet
};

# default locate of the config file
my($configfile) = "rrdgraphing.conf";

# check for command line arguments
my(%opts);
my($ret) = getopts('hvqVc:', \%opts);

# if invalid argument given, $ret will not be 1
&usage() if $ret != 1;

# first process the arguments which might mean we exit now

# -h is usage
if($opts{h}) {
    &usage();
}
# -V is version
if($opts{V}) {
    print "graph.pl version: $version\n";
    exit(1);
}

# Then try getting the config

# -c specifies the config file location
if($opts{c}) {
    $configfile = $opts{c};
}
# suck in the config
&log("reading config from $configfile\n");
do $configfile;

# Then any options we might want to override the config with

# -v is verbose
if($opts{v}) {
    $verbose = $opts{v};
}
# -q is verbose
if($opts{q}) {
    $quiet = $opts{q};
    # if we're meant to be quiet, we can hardly be verbose!
    $verbose = 0;
}

# Read the contents of the base directory
# and pull out the list of subdirectories (except . and .. :)
opendir(DIR, $rrddir);
my(@rrddirlist) = grep { -d "$rrddir/$_" && !/^\.$/ && !/^\.\.$/ } readdir(DIR);
closedir DIR;

# look through each directoty, as they might
# contain rrds for a particular machine
foreach my $machine (@rrddirlist) {
    # Read the contents of the directory
    opendir(DIR, "$rrddir/$machine");
    my(@rrdlist) = grep { /\.rrd$/ && -f "$rrddir/$machine/$_" } readdir(DIR);
    closedir DIR;
    
    # See what rrd we have, and generate the graphs accordingly
    foreach my $rrd (@rrdlist) {
        chomp $rrd;
        # stat the file
        my($dev,$ino,$mode,$nlink,$uid,$gid,$rdev,$size,$atime,$mtime,
         $ctime,$blksize,$blocks) = stat("$rrddir/$machine/$rrd");
        # check if it's old enough to be deleted
        if((time - $mtime) > $maxrrdage) {
            # do we delete the rrd, or just ignore it?
            if($deleterrds) {
                # if so, delete it
                unlink("$rrddir/$machine/$rrd");
                &log("deleted old rrd $rrddir/$machine/$rrd\n");
            }
            else {
                &log("ignored old rrd $rrddir/$machine/$rrd\n");
            }
            # no more processing required for this rrd
            next;
        }
        if($rrd =~ /^(cpu)\.rrd$/) {
            my(@data);
            my(@rawdata);
            push @data, "LINE2:$1:idle:idle#00FF00:OK:idle cpu  ";
            push @data, "LINE2:$1:user:user#0000FF:OK:user cpu  ";
            push @data, "LINE2:$1:kernel:kernel#00FFFF:OK:kernel cpu";
            push @data, "LINE2:$1:swap:swap#FF00FF:OK:swap cpu  ";
            push @data, "LINE2:$1:iowait:iowait#FF0000:OK:iowait cpu";
            push @rawdata, "--upper-limit=100";
            &makegraph($machine, $1, "CPU Usage for $machine", "% cpu time", \@data, \@rawdata);
        }
        if($rrd =~ /^(mem)\.rrd$/) {
            my(@data);
            my(@rawdata);
            # we don't actually want to display free or total memory,
            # although we need it to work out peruse...
            push @data, "NONE:$1:free:free#CCCCFF:NONE:free memory";
            push @data, "NONE:$1:total:total#0000FF:NONE:total memory\\n";
            # calculate peruse - note that we only use 'free' if it's less than total
            # (this is to avoid negative percentages :)
            push @rawdata, "CDEF:peruse=total,free,total,LT,free,0,IF,-,total,/,100,*";
            # and add it to the graph
            push @rawdata, "AREA:peruse#CCCCFF:% memory in use";
            push @rawdata, "--upper-limit=100";
            push @rawdata, "--base=1024";
            # add some nice values to the legend
            &addlegend(\@rawdata, "peruse");
            # put the total memory on the graph so we can map percentages to real values
            push @rawdata, "GPRINT:total:LAST:Current total memory\\: \%.2lf %sb\\c";
            &makegraph($machine, $1, "Memory Usage for $machine", "% memory in use", \@data, \@rawdata);
        }
        if($rrd =~ /^(load)\.rrd$/) {
            my(@data);
            push @data, "LINE2:$1:load1:load1#CCCCFF:OK: 1 min load average";
            push @data, "LINE2:$1:load5:load5#7777FF:OK: 5 min load average";
            push @data, "LINE2:$1:load15:load15#0000FF:OK:15 min load average";
            &makegraph($machine, $1, "Loads for $machine", "load average", \@data);
        }
        if($rrd =~ /^(proc)\.rrd$/) {
            my(@data);
            push @data, "LINE2:$1:cpu:cpu#00FF00:OK:cpu processes     ";
            push @data, "LINE2:$1:sleeping:sleeping#0000FF:OK:sleeping processes";
            push @data, "LINE2:$1:stopped:stopped#00FFFF:OK:stopped processes ";
            push @data, "LINE2:$1:total:total#FF00FF:OK:total processes   ";
            push @data, "LINE2:$1:zombie:zombie#FF0000:OK:zombie processes  ";
            &makegraph($machine, $1, "Processes on $machine", "no. of processes", \@data);
        }
        if($rrd =~ /^(swap)\.rrd$/) {
            my(@data);
            my(@rawdata);
            # we don't actually want to display free or total swap,
            # although we need it to work out peruse...
            push @data, "NONE:$1:free:free#CCCCFF:NONE:free swap";
            push @data, "NONE:$1:total:total#0000FF:NONE:total swap\\n";
            # calculate peruse - note that we only use 'free' if it's less than total
            # (this is to avoid negative percentages :)
            push @rawdata, "CDEF:peruse=total,free,total,LT,free,0,IF,-,total,/,100,*";
            # and add it to the graph
            push @rawdata, "AREA:peruse#CCCCFF:% swap in use";
            push @rawdata, "--upper-limit=100";
            push @rawdata, "--base=1024";
            # add some nice values to the legend
            &addlegend(\@rawdata, "peruse");
            # put the total swap on the graph so we can map percentages to real values
            push @rawdata, "GPRINT:total:LAST:Current total swap\\: \%.2lf %sb\\c";
            &makegraph($machine, $1, "Swap Usage for $machine", "% swap in use", \@data, \@rawdata);
        }
        if($rrd =~ /^(users)\.rrd$/) {
            my(@data);
            push @data, "AREA:$1:count:count#CCCCFF:OK:user count";
            &makegraph($machine, $1, "User Count for $machine", "no. of users", \@data);
        }
        if($rrd =~ /^(paging)\.rrd$/) {
            my(@data);
            push @data, "LINE2:$1:pageins:pageins#00FF00:OK:pages paged in ";
            push @data, "LINE2:$1:pageouts:pageouts#0000FF:OK:pages paged out";
            &makegraph($machine, $1, "Paging on $machine", "pages per second", \@data);
        }
        if($rrd =~ /^(disk)-(\S+)\.rrd$/) {
            my(@data);
            my(@rawdata);
            # we need this lot for our calculations, but we'll never show them
            push @data, "NONE:$1-$2:total:total#0000FF:NONE:total size\\n";
            push @data, "NONE:$1-$2:used:used#CCCCFF:NONE:used space";
            push @data, "NONE:$1-$2:totalinodes:totalinodes#000000:NONE:total inodes";
            push @data, "NONE:$1-$2:freeinodes:freeinodes#000000:NONE:free inodes";
            # calculate peruse, add it to the graph, and add a legend
            push @rawdata, "CDEF:peruse=used,total,/,100,*";
            push @rawdata, "AREA:peruse#CCCCFF:% disk used  ";
            &addlegend(\@rawdata, "peruse");
            # put the total space on the graph so we can map percentages to real values
            push @rawdata, "GPRINT:total:LAST:Current total space\\: \%.2lf %sb\\c";
            # calculate perinodeuse, add it to the graph, and add a legend
            push @rawdata, "CDEF:perinodeuse=totalinodes,freeinodes,totalinodes,LT,freeinodes,0,IF,-,totalinodes,/,100,*";
            push @rawdata, "LINE2:perinodeuse#FF4444:% inodes used";
            push @rawdata, "--upper-limit=100";
            push @rawdata, "--base=1024";
            &addlegend(\@rawdata, "perinodeuse");
            # put the total inodes on the graph so we can map percentages to real values
            push @rawdata, "GPRINT:totalinodes:LAST:Current total inodes\\: \%.2lf %s\\c";
            # some name tidting
            my($type) = $1;
            my($name) = $2;
            my($nicename) = $2;
            $nicename =~ s/$hex_slash/\//g;
            $nicename =~ s/$hex_underscore/_/g;
            &makegraph($machine, "$type-$name", "Disk Usage for $machine on $nicename", "% usage", \@data, \@rawdata);
        }
        if($rrd =~ /^(diskio)-(\S+)\.rrd$/) {
            my(@data);
            my(@rawdata);
            push @data, "LINE2:$1-$2:rbytes:rbytes#00FF00:OK:read bytes ";
            push @data, "LINE2:$1-$2:wbytes:wbytes#0000FF:OK:write bytes";
            push @rawdata, "--base=1024";
            &makegraph($machine, "$1-$2", "Disk IO for $machine on $2", "bytes per second", \@data, \@rawdata);
        }
        if($rrd =~ /^(net)-(\S+)\.rrd$/) {
            my(@data);
            my(@rawdata);
            push @data, "LINE2:$1-$2:rx:rx#00FF00:OK:received bytes  ";
            push @data, "LINE2:$1-$2:tx:tx#0000FF:OK:transfered bytes";
            push @rawdata, "--base=1024";
            &makegraph($machine, "$1-$2", "Network IO for $machine on $2", "bytes per second", \@data, \@rawdata);
        }
        # probably a queue with a name like this :)
        if($rrd =~ /^(\d+)_0\.rrd$/) {
            my(@data);
            my(@rawdata);
            my($baserrd) = $1;
            my($i) = 0;
            while( -f "$rrddir/$machine/$baserrd\_$i.rrd" ) {
                push @data, "LINE2:$baserrd\_$i:size:size$i" . &get_colour($i) . ":OK:queue$i size ";
                ++$i;
            }
            push @data, "LINE2:$baserrd\_0:total:total#FF0000:OK:packets/sec ";
            my($comment);
            if(-f "$rrddir/$machine/$baserrd.def") {
                open(DEF, "$rrddir/$machine/$baserrd.def");
                $comment = <DEF>;
                chomp $comment if defined $comment;
            }
            $comment = "unknown queue" if not defined $comment;
            &makegraph($machine, $baserrd, $comment, "", \@data, \@rawdata);
        }
    }
    # have a last check, maybe we can remove the directory now?
    # (only if we're deleting stuff)
    if($deleterrds) {
        # Read the contents of the directory
        opendir(DIR, "$rrddir/$machine");
        my(@dirlist) = grep { !/^\.$/ && !/^\.\.$/ } readdir(DIR);
        closedir DIR;
        if($#dirlist == -1) {
            rmdir "$rrddir/$machine";
            &log("deleting empty rrd directory $rrddir/$machine\n");
        }
    }
}

if($deleteimgs) {
    # Read the contents of the graphs directory
    # and pull out the list of subdirectories (except . and .. :)
    opendir(DIR, $imgdir);
    my(@imgdirlist) = grep { -d "$imgdir/$_" && !/^\.$/ && !/^\.\.$/ } readdir(DIR);
    closedir DIR;

    # look through each directoty, as they might
    # contain images for a particular machine
    foreach my $machine (@imgdirlist) {
        # Read the contents of the directory
        opendir(DIR, "$imgdir/$machine");
        my(@imglist) = grep { /\.png$/ && -f "$imgdir/$machine/$_" } readdir(DIR);
        closedir DIR;

        # See what rrd we have, and generate the graphs accordingly
        foreach my $img (@imglist) {
            chomp $img;
            # stat the img
            my($dev,$ino,$mode,$nlink,$uid,$gid,$rdev,$size,$atime,$mtime,
             $ctime,$blksize,$blocks) = stat("$imgdir/$machine/$img");
            # check if it's old enough to be deleted
            if((time - $mtime) > $maximgage) {
                # if so, delete it
                unlink("$imgdir/$machine/$img");
                &log("deleted old image $imgdir/$machine/$img\n");
            }
        }
        # have a last check, maybe we can remove the directory now?
        # Read the contents of the directory
        opendir(DIR, "$imgdir/$machine");
        my(@dirlist) = grep { !/^\.$/ && !/^\.\.$/ } readdir(DIR);
        closedir DIR;
        if($#dirlist == -1) {
            rmdir "$imgdir/$machine";
            &log("deleted empty image directory $imgdir/$machine\n");
        }
    }
}

exit(0);


#
# subroutine to make some graphs
#
# $machine   = name of the machine
#              (eg. kernow.ukc.ac.uk)
# $type      = the type of graph for the machine
#              (eg. cpu)
# $title     = the title for the graph
#              (eg. kernow CPU usage)
# $vlabel    = the vertical label to apply to the left side of the graph
#              (eg. kb/s)
# $dataref   = a reference to an array containing information for the graph
#              elements of format: "gtype:rrdname:dsname:name#colour:legend:comment with spaces"
#              (if gtype is "NONE" only a DEF of 'name' will be defined, no line will be plotted)
#              (if legend is "NONE" the latest/average/max/min legend won't be printed)
# $rawcmdref = a reference to an array containing raw rrd commands
#              elements a single command each, no spaces
#

sub makegraph() {
    my($machine, $type, $title, $vlabel, $dataref, $rawcmdref) = @_;
    # pass in these arrays by reference
    my(@data) = @$dataref if defined $dataref;
    my(@rawcmd) = @$rawcmdref if defined $rawcmdref;
    # check if directory exists for images
    if(! -d "$imgdir/$machine") {
        # not sure on this umask, but it seems to work?
        mkdir "$imgdir/$machine", 0777;
        &log("created directory $imgdir/$machine\n");
    }
    my(@rrdcmd);
    foreach my $dataitem (@data) {
        # dataitem should be: "gtype:rrdname:dsname:name#colour:legend:comment with spaces"
        # (if gtype is "NONE" only a DEF of 'name' will be defined, no line will be plotted)
        # (if legend is "NONE" the latest/average/max/min legend won't be printed)
        if($dataitem =~ /^(\S+):(\S+):(\S+):(\S+)#(.{6}):(\S+):(.*)$/) {
            push @rrdcmd, "DEF:$4=$rrddir/$machine/$2.rrd:$3:AVERAGE";
            if($1 ne "NONE") {
                push @rrdcmd, "$1:$4#$5:$7";
                if($6 ne "NONE") {
                    # add some nice values to the legend
                    &addlegend(\@rrdcmd, $4);
                }
            }
        }
    }
    push @rrdcmd, "--title=$title";
    push @rrdcmd, "--imgformat=PNG";
    push @rrdcmd, "--lower-limit=0";
    push @rrdcmd, "--vertical-label=$vlabel";
    # not entirely convinced this is good...
    push @rrdcmd, "--alt-autoscale-max";
    # add any further raw commands
    push @rrdcmd, @rawcmd;
    RRDs::graph ("$imgdir/$machine/$type-3h.png", "--start=-10800", @rrdcmd);
    my($err_3h) = RRDs::error;
    &log("created $imgdir/$machine/$type-3h.png\n") unless $err_3h;
    &error("Error generating 3h graph for $machine/$type: $err_3h\n") if $err_3h;
    RRDs::graph ("$imgdir/$machine/$type-1d.png", "--start=-86400", @rrdcmd);
    my($err_1d) = RRDs::error;
    &log("created $imgdir/$machine/$type-1d.png\n") unless $err_1d;
    &error("Error generating 1d graph for $machine/$type: $err_1d\n") if $err_1d;
    RRDs::graph ("$imgdir/$machine/$type-1w.png", "--start=-604800", @rrdcmd);
    my($err_1w) = RRDs::error;
    &log("created $imgdir/$machine/$type-1w.png\n") unless $err_1w;
    &error("Error generating 1w graph for $machine/$type: $err_1w\n") if $err_1w;
    RRDs::graph ("$imgdir/$machine/$type-1m.png", "--start=-2678400", @rrdcmd);
    my($err_1m) = RRDs::error;
    &log("created $imgdir/$machine/$type-1m.png\n") unless $err_1m;
    &error("Error generating 1m graph for $machine/$type: $err_1m\n") if $err_1m;
    RRDs::graph ("$imgdir/$machine/$type-1y.png", "--start=-31536000", @rrdcmd);
    my($err_1y) = RRDs::error;
    &log("created $imgdir/$machine/$type-1y.png\n") unless $err_1y;
    &error("Error generating 1y graph for $machine/$type: $err_1y\n") if $err_1y;
    return;
}

# subroutine to add a legend
# accepts reference to an array and a name
sub addlegend() {
    my($dataref, $name) = @_;
    push @$dataref, "GPRINT:$name:LAST:Current\\: \%8.2lf %s";
    push @$dataref, "GPRINT:$name:AVERAGE:Average\\: \%8.2lf %s";
    push @$dataref, "GPRINT:$name:MAX:Max\\: \%8.2lf %s\\n";
}

# hacky subroutine to return a colour
# could be done much better somehow :/
sub get_colour {
    my($col) = @_;
    if($col == 0) {
        return "#0000FF";
    }
    elsif($col == 1) {
        return "#00FF00";
    }
    elsif($col == 2) {
        return "#FF00FF";
    }
    elsif($col == 3) {
        return "#FFFF00";
    }
    elsif($col == 4) {
        return "#00FFFF";
    }
    else {
        return "#000066";
    }
}

# prints out usage information then exits
sub usage() {
    print "Usage: graph.pl [options]\n";
    print "Options\n";
    print "  -c config        Specifies the configuration file\n";
    print "                    default: rrdgraphing.conf\n";
    print "  -v               Be verbose about what's happening\n";
    print "  -q               Be quiet, even supress errors\n";
    print "  -V               Print version number\n"; 
    print "  -h               Prints this help page\n";
    exit(1);
}       

# prints a log message if verbose is turned on
sub log() {
    my($msg) = @_;
    print $msg if $verbose;
}

# prints an error message unless quiet is turned on
sub error() {
    my($msg) = @_;
    print STDERR $msg unless $quiet;
}
